from collections import defaultdict
from typing import Dict, Set, Tuple, Union
from cse355_machine_design.automata.base import _Automata
from itertools import product
from cse355_machine_design.errors import DetailedError
from cse355_machine_design import registry


class _DFA(_Automata):
    _transition_table: Dict[Tuple[str, str], str]
    _final_states: Set[str]

    def __init__(
        self,
        Q: Set[str],
        Sigma: Set[str],
        delta: Dict[Tuple, str],
        q0: str,
        F: Set[str],
    ) -> None:
        _delta: Dict[Tuple[str, str], str] = delta  # type: ignore
        self._states = Q
        self._input_symbols = Sigma
        self._transition_table = _delta
        self._start_state = q0
        self._final_states = F
        super().__init__()

    def verify_legality(self) -> None:
        transitions_valid = True
        seen_transition: set[Tuple[str, str]] = set()
        err_str = ""
        for k, v in self._transition_table.items():
            transition_current_is_state = k[0] in self._states
            transition_final_is_state = v in self._states
            transition_input_is_alphabet = k[1] in self._input_symbols
            if k in seen_transition:
                err_str += f'delta has an attempt to redefine the transition from "{k[0]}" while reading input "{k[1]}"\n'
            else:
                seen_transition.add(k)
            if not transition_current_is_state:
                err_str += f'delta has a transition from "{k[0]}" on input "{k[1]}" to "{v}", but "{k[0]}" is not in Q\n'
            if not transition_input_is_alphabet:
                err_str += f'delta has a transition from "{k[0]}" on input "{k[1]}" to "{v}", but "{k[1]}" is not in Sigma\n'
            if not transition_final_is_state:
                err_str += f'delta has a transition from "{k[0]}" on input "{k[1]}" to "{v}", but "{v}" is not in Q\n'

            transitions_valid &= (
                transition_current_is_state
                and transition_input_is_alphabet
                and transition_final_is_state
            )

        if not transitions_valid:
            raise DetailedError(
                "(Q, Sigma and/or delta are incorrectly defined)", err_str
            )

        err_str = ""

        all_entries = set(product(self._states, self._input_symbols))

        # check all entries to make sure transition table is complete
        transitions_complete = True
        for necessary_transition in all_entries:
            transition_found = necessary_transition in self._transition_table.keys()
            if not transition_found:
                err_str += f'Based on Q,Sigma, we need a transition from "{necessary_transition[0]}" to some other state on input symbol "{necessary_transition[1]}", but this transition was not found\n'
            transitions_complete &= transition_found

        if not transitions_complete:
            raise DetailedError(
                "(Based on the definitions of Q and Sigma, it is the case that delta is missing transitions)",
                err_str,
            )

        start_valid = self._start_state in self._states
        if not start_valid:
            raise DetailedError(
                "(Start state q_0 not found in Q)",
                f'"{self._start_state}" was not found in Q',
            )

        err_str = ""
        final_valid = True
        for potential_state in self._final_states:
            potential_state_valid = potential_state in self._states
            if not potential_state_valid:
                err_str += f'"{potential_state}" was declared as a final state, but it is not found in Q\n'
            final_valid &= potential_state_valid
        final_valid = self._final_states.issubset(self._states)
        if not final_valid:
            raise DetailedError("(F is not a subset of Q)", err_str)

    def evaluate_one_step(
        self, input_char: str, current_state: str, enable_trace=False
    ) -> str:
        new_state = self._transition_table[(current_state, input_char)]
        if enable_trace:
            print(
                f'Reading input "{input_char}". This causes us to transition from "{current_state}" to "{new_state}"'
            )
        return new_state

    def evaluate(self, input_str: str, enable_trace=False) -> Union[bool, None]:
        if enable_trace:
            print(f'Starting at state "{self._start_state}"')
        current_state = self._start_state
        for input_char in input_str:
            current_state = self.evaluate_one_step(
                input_char, current_state, enable_trace
            )
        is_final = current_state in self._final_states
        accept_str = ("", "accept") if is_final else (" not", "reject")
        if enable_trace:
            print(
                f'Finished reading input. We are now in state "{current_state}". This is{accept_str[0]} a final state, so we {accept_str[1]}'
            )
        return current_state in self._final_states

    def _generate_dot_string(self) -> str:
        res = 'digraph {rankdir="LR";ranksep=0.2;edge[minlen=3];'
        res += 'subgraph cluster {rank=same;edge[minlen=default];rankdir="LR";peripheries=0;'
        res += 'n0 [label="", shape=none, width=0, height=0];'
        res += '"{}" [shape={}];'.format(
            self._start_state,
            "doublecircle" if self._start_state in self._final_states else "circle",
        )
        res += 'n0 -> "{}"'.format(self._start_state)
        res += "};"
        for state in self._states:
            if state == self._start_state:
                continue
            res += '"{}" [shape={}];'.format(
                state, "doublecircle" if state in self._final_states else "circle"
            )
        tt_restructured: defaultdict[Tuple[str, str], list[str]] = defaultdict(list)

        for (from_state, input_symbol), to_state in self._transition_table.items():
            tt_restructured[(from_state, to_state)].append(input_symbol)

        for (from_state, to_state), transition_list in tt_restructured.items():
            label = "{}".format(transition_list[0])
            for i, input_symbol in enumerate(transition_list):
                if i == 0:
                    continue
                label += ","
                if i % 3 == 2:
                    label += "\n"
                label += "{}".format(transition_list[i])
            res += '"{}" -> "{}" [label="{}"];'.format(from_state, to_state, label)

        res += "}"
        return res

    def submit_as_answer(self, question_number: int):
        registry.add_to_registry("dfa", question_number, self)



class DetailedError(RuntimeError):
    def __init__(self, general_message: str, detailed_message: str) -> None:
        err_str = f"\n\nv~~~~ See below for detailed error ~~~v {general_message}\n\n{detailed_message}"
        super().__init__(err_str)



from __future__ import annotations
from typing import TYPE_CHECKING, Union
import json

if TYPE_CHECKING:
    from typing import Dict, TypedDict, KeysView, ValuesView

    from .. import automata
    from ..automata import base as automata_base

    class _Registry(TypedDict):
        dfa: Dict[int, automata.DFA]
        nfa: Dict[int, automata.NFA]

else:
    from .. import automata


_registry: _Registry = {"dfa": {}, "nfa": {}}


def add_to_registry(
    automata_type: str, question_number: int, submission: automata_base._Automata
) -> None:
    _registry[automata_type][question_number] = submission


def get_dfa(question_number: int) -> Union[automata.DFA, None]:
    return _registry["dfa"].get(question_number)


def get_nfa(question_number: int) -> Union[automata.NFA, None]:
    return _registry["nfa"].get(question_number)


def export_submissions() -> None:
    mod_registry = {}
    for outer_k, outer_v in _registry.items():
        mod_registry[outer_k] = {}
        for inner_k, inner_v in outer_v.items():  # type: ignore
            mod_registry[outer_k][inner_k] = inner_v.export_to_dict()
    out_str = json.dumps(mod_registry)
    with open("submissions.json", "w") as out:
        out.write(out_str)


def import_submissions(path: str = "submissions.json") -> None:
    with open(path, "r") as in_file:
        mod_registry = json.load(in_file)

    dfas: Dict[str, Dict] = mod_registry["dfa"]
    for submission_num, dfa_desc in dfas.items():
        Q = set(dfa_desc["states"])
        Sigma = set(dfa_desc["input_symbols"])
        q0 = dfa_desc["start_state"]
        F = set(dfa_desc["finals"])
        delta = {}
        for transition in dfa_desc["delta"]:
            delta[(transition["from"], transition["input"])] = transition["to"][0]
        dfa = automata.DFA(Q, Sigma, delta, q0, F)
        add_to_registry(dfa_desc["type"], int(submission_num), dfa)

    nfas: Dict[str, Dict] = mod_registry["nfa"]
    for submission_num, nfa_desc in nfas.items():
        Q = set(nfa_desc["states"])
        Sigma = set(nfa_desc["input_symbols"])
        q0 = nfa_desc["start_state"]
        F = set(nfa_desc["finals"])
        delta = {}
        for transition in nfa_desc["delta"]:
            delta[(transition["from"], transition["input"])] = set(transition["to"])
        nfa = automata.NFA(Q, Sigma, delta, q0, F, nfa_desc["epsilon_char"])
        add_to_registry(nfa_desc["type"], int(submission_num), nfa)



from abc import ABC, abstractmethod
from typing import Set, Union

# from PIL import Image, ImageOps
# import matplotlib.pyplot as plt
# import numpy as np
# import urllib.request
# from urllib.parse import quote
# import io
import render_html
import pathlib


class _Automata(ABC):
    _states: Set[str]
    _input_symbols: Set[str]
    _start_state: str

    def __init__(self) -> None:
        self.verify_legality()

    @abstractmethod
    def verify_legality(self) -> None:
        """
        Verify the basic n-tuple requirements
        """
        raise NotImplementedError("Abstract method not callable")

    @abstractmethod
    def evaluate(self, input_str: str, enable_trace=0) -> Union[bool, None]:
        """
        Evaluate the automata on a string
        """
        raise NotImplementedError("Abstract method not callable")

    @abstractmethod
    def submit_as_answer(self, question_number: int) -> None:
        """
        Submit this automata as the answer to a given question.

        #### Enter the question number exactly as seen on the homework assignment.

        This overrides any prior calls to this function for the given question.
        """
        raise NotImplementedError("Abstract method not callable")

    @abstractmethod
    def export_to_dict(self) -> dict:
        """
        In the following Python Dict spec, describe the machine:
        {
            "type": "dfa",
            "states": str[],
            "input_symbols": str[]
            "delta": {
                        "from": str,
                        "to": str,
                        "input": str[] //<-- singleton array
                    }[]
            "start_state": str,
            "finals": str[]
        } |
        {
            "type": "nfa",
            "states": str[],
            "input_symbols": str[]
            "delta": {
                        "from": str,
                        "to": str[],
                        "input": str
                    }[]
            "start_state": str,
            "finals": str[],
            "epsilon_char": str
        }
        """
        raise NotImplementedError("Abstract method not callable")

    @abstractmethod
    def _generate_dot_string(self) -> str:
        """
        Display the machine state diagram as a graphviz plot
        """
        raise NotImplementedError("Abstract method not callable")

    def display_state_diagram(self) -> None:
        html_contents = (
            '<!doctypehtml><title>Automata Visualization</title><meta content="View your automata in the browser"name=description><script src=https://cdn.jsdelivr.net/npm/@viz-js/viz@3.2.4/lib/viz-standalone.min.js></script><script>Viz.instance().then(function(e){var n=e.renderSVGElement(\''
            + self._generate_dot_string()
            + '\');document.getElementById("graph").appendChild(n)})</script><div id=graph></div>'
        )
        render_html.render_in_browser(
            html_contents, str(pathlib.Path.cwd() / "preview.html")
        )
